#include "evolution.h"

#define NUM_VERTS 5
int NUM_POLYS=50;
int algorithm = 0;
int verbose = 0;

struct polygon {
    CvPoint points[NUM_VERTS];
    CvScalar colour;
    float alpha;
};

struct polygon* polygons;
struct polygon* trialpolys;

long int calc_fitness(IplImage* img1, IplImage* img2) {
    long int tempfitness = 0;
    int x, y;

    for(y = 0; y < img1->height; y++) {
        uchar* ptr1 = img1->imageData + y*img1->widthStep;
        uchar* ptr2 = img2->imageData + y*img2->widthStep;
        for(x = 0; x < img1->width; x++) {
            tempfitness += abs(ptr1[x*3] - ptr2[x*3]);
            tempfitness += abs(ptr1[x*3+1] - ptr2[x*3+1]);
            tempfitness += abs(ptr1[x*3+2] - ptr2[x*3+2]);
        }
    }
    return tempfitness;
}

void AlphaBlending(IplImage *pImg1, float alpha, IplImage *pImg2, float beta, IplImage *pDest, IplImage *pMask)
{
   // pMask must have 1 channel
    int x, y;

   for(y = 0; y<pImg1->height; y++ )
   {
      uchar* ptrImg1 = (uchar*) (pImg1->imageData + y * pImg1->widthStep);
      uchar* ptrImg2 = (uchar*) (pImg2->imageData + y * pImg2->widthStep);
      uchar* ptrDest = (uchar*) (pDest->imageData + y * pDest->widthStep);
      uchar* ptrMask = (uchar*) (pMask->imageData + y * pMask->widthStep);
       
      for(x = 0; x<pImg1->width; x++)
      {
         if(ptrMask[x] > 0)
         {
            ptrDest[3*x] = alpha*ptrImg1[3*x] + beta*ptrImg2[3*x];
            ptrDest[3*x+1] = alpha*ptrImg1[3*x+1] + beta*ptrImg2[3*x+1];
            ptrDest[3*x+2] = alpha*ptrImg1[3*x+2] + beta*ptrImg2[3*x+2];
         }
      }
   }
} 

void mutate(struct polygon* poly, int width, int height) {
    if(!algorithm) {
        // Soft mutate
        int r = rand()%2;
        if(r) {
            // Modify single point
            int point = rand()%NUM_VERTS;
            poly->points[point] = cvPoint(rand()%width, rand()%height);
        } else {
            // Modify colour
            poly->colour = CV_RGB(rand()%256, rand()%256, rand()%256);
            poly->alpha = ((float)(rand()%256))/255.0;
        }
    } else {
        // Hard mutate
        int i;
        for(i = 0; i<NUM_VERTS; i++) {
            poly->points[i] = cvPoint(rand()%width, rand()%height);
        }
        poly->colour = CV_RGB(rand()%256, rand()%256, rand()%256);
        poly->alpha = ((float)(rand()%256))/255.0;
    }
}

void getargs(int argc, char** argv, char** image) {
    int i;

    for(i=0; i<argc; i++) {
        if(!strcmp(argv[i], "-p")) {
            NUM_POLYS = atoi(argv[++i]);
        }
        if(!strcmp(argv[i], "-a")) {
            if(!strcmp(argv[i+1], "soft")) {
                algorithm = 0;
            } else if(!strcmp(argv[i+1], "hard")) {
                algorithm = 1;
            }
        }
        if(!strcmp(argv[i], "-v")) {
            verbose = 1;
        }
        *image = argv[i];
    }
}

int main(int argc, char** argv) {
    long int fitness = LONG_MAX;
    long int newfitness = 0;
    int i, j;
    int count = 0;
    char c;
    char* image;

    if(argc<2) {
        printf( "Usage: %s [OPTIONS] <source file>\n"
                "\n"
                "OPTIONS\n"
                "   -v              Verbose mode\n"
                "   -p N            Generate image using N polygons. Default is 50.\n"
                "   -a [hard|soft]  Use hard or soft mutation algorithm. Default is soft.\n", argv[0]);
        return 1;
    }

    getargs(argc, argv, &image);

    if(algorithm) {
        printf("Hard mutation algorithm selected\n");
    }

    printf("%d polygons, each with %d points\n", NUM_POLYS, NUM_VERTS);

    trialpolys = malloc(sizeof(struct polygon)*NUM_POLYS);
    polygons = malloc(sizeof(struct polygon)*NUM_POLYS);

    IplImage* source = cvLoadImage(image, CV_LOAD_IMAGE_COLOR);
    if(!source) {
        printf("Could not open source image \"%s\" for reading. Aborting.\n", image);
        return 1;
    }
    IplImage* dest = cvCreateImage(cvSize(source->width, source->height), IPL_DEPTH_8U, 3);
    IplImage* trial = cvCreateImage(cvSize(source->width, source->height), IPL_DEPTH_8U, 3);
    IplImage* shape = cvCreateImage(cvSize(source->width, source->height), IPL_DEPTH_8U, 3);
    IplImage* mask = cvCreateImage(cvSize(source->width, source->height), IPL_DEPTH_8U, 1);

    cvSetZero(trial);
    cvSetZero(shape);

    cvNamedWindow("Source", CV_WINDOW_AUTOSIZE);
    cvShowImage("Source", source);

    cvNamedWindow("Trial", CV_WINDOW_AUTOSIZE);
    cvNamedWindow("Approximation", CV_WINDOW_AUTOSIZE);

    srand((unsigned)time(NULL));

    printf("Process begun, press escape to exit.");

    for(i=0; i<NUM_POLYS; i++) {
        for(j=0; j<NUM_VERTS; j++) {
            polygons[i].points[j] = cvPoint(rand()%source->width, rand()%source->height);
        }
        polygons[i].colour = CV_RGB(rand()%256, rand()%256, rand()%256);
        polygons[i].alpha = ((float)(rand()%256))/255.0;
    }
    memcpy(trialpolys, polygons, sizeof(struct polygon)*NUM_POLYS); 

    while(1) {
        mutate(&trialpolys[rand()%NUM_POLYS], source->width, source->height);
        for(i=0; i<NUM_POLYS; i++) {
            cvFillConvexPoly(shape, trialpolys[i].points, NUM_VERTS, trialpolys[i].colour, 4, 0);
            cvConvertImage(shape, mask, 0);
            AlphaBlending(shape, trialpolys[i].alpha, trial, 1-trialpolys[i].alpha, trial, mask);
            cvSetZero(shape);
        }
        newfitness = calc_fitness(trial, source);
        if(newfitness < fitness) {
            fitness = newfitness;
            memcpy(polygons, trialpolys, sizeof(struct polygon)*NUM_POLYS);
            if(verbose)
                printf("Count: %d, New fitness: %ld\n", ++count, fitness);
            cvCopy(trial, dest, NULL);
            cvShowImage("Approximation", dest);
        }
        memcpy(trialpolys, polygons, sizeof(struct polygon)*NUM_POLYS);
        cvShowImage("Trial", trial);
        cvSetZero(trial);
        c = cvWaitKey(2);
        if(c==27) // Escape
            break;
    }

    cvReleaseImage(&source);
    cvReleaseImage(&dest);
    cvReleaseImage(&trial);
    cvReleaseImage(&shape);
    cvReleaseImage(&mask);

    cvDestroyWindow("Source");
    cvDestroyWindow("Trial");

}
